###############################################################################
#
# The gcc versioning information.  In a sed command below, the specfile winds
# pre-release version numbers in BASE-VER back to the last actually-released
# number.
%global DATE 20190507
%global SVNREV 270976
%global gcc_version 8.3.1
%global gcc_major 8
%global gcc_release 3

%global bpf_binutils_version 2.30-1
%global isl_version 0.16.1
%global isl_libmajor 15

%global _performance_build 1
# Hardening slows the compiler way too much.
%undefine _hardened_build
%if 0%{?fedora} > 27 || 0%{?rhel} >= 8
# Until annobin is fixed (#1519165).
%undefine _annotated_build
%endif

Summary: BPF C compiler
Name: bpf-gcc
Version: %{gcc_version}
Release: %{gcc_release}%{?dist}
# libgcc, libgfortran, libmudflap, libgomp, libstdc++ and crtstuff have
# GCC Runtime Exception.
License: GPLv3+ and GPLv3+ with exceptions and GPLv2+ with exceptions and LGPLv2+ and BSD
URL: http://gcc.gnu.org
BuildRequires:  gcc-c++
#BuildRequires: isl-devel >= %{isl_version}

# The source for this package was pulled from upstream's vcs.  Use the
# following commands to generate the tarball:
# svn export svn://gcc.gnu.org/svn/gcc/branches/redhat/gcc-8-branch@%%{SVNREV} gcc-%%{version}-%%{DATE}
# tar cf - gcc-%%{version}-%%{DATE} | xz -9e > gcc-%%{version}-%%{DATE}.tar.xz
%global srcdir gcc-%{version}-%{DATE}
Source0: %{srcdir}.tar.xz

Patch0: gcc8-hack.patch
Patch2: gcc8-i386-libgomp.patch
Patch3: gcc8-sparc-config-detection.patch
Patch4: gcc8-libgomp-omp_h-multilib.patch
Patch5: gcc8-libtool-no-rpath.patch
Patch6: gcc8-isl-dl.patch
Patch7: gcc8-libstdc++-docs.patch
Patch8: gcc8-no-add-needed.patch
Patch9: gcc8-foffload-default.patch
Patch10: gcc8-Wno-format-security.patch
Patch11: gcc8-rh1512529-aarch64.patch
Patch12: gcc8-mcet.patch
Patch13: gcc8-rh1574936.patch
Patch14: gcc8-libgcc-hardened.patch
Patch15: gcc8-rh1670535.patch
Patch16: gcc8-pr85400.patch
Patch17: gcc8-libgomp-20190503.patch
Patch18: gcc8-pr86747.patch
Patch19: gcc8-libgomp-testsuite.patch
Patch20: gcc8-rh1711346.patch
Patch21: gcc8-rh1730380.patch
Patch22: gcc8-pr86098.patch
Patch23: gcc8-pr90139.patch
Patch24: gcc8-pr90756.patch

Patch30: gcc8-rh1668903-1.patch
Patch31: gcc8-rh1668903-2.patch
Patch32: gcc8-rh1668903-3.patch

# Patches from Ampere for aarch64. See
# https://git.theobroma-systems.com/ampere-computing/gcc.git/log/?h=gcc-8_2_0-amp-branch.
Patch2001: amp-1-aarch64-retpoline.patch
Patch2002: amp-2-revert-2017-10-24-richard.patch
Patch2003: amp-3-aarch64-correct-shift-amount.patch
Patch2004: amp-4-aarch64-cpu-support-ampere.patch
Patch2005: amp-5-aarch64-emag-xgene-cost-model.patch
Patch2006: amp-6-update-cost-table-for-xgene1.patch
Patch2007: amp-7-aarch64-add-xgene1-prefetch-tuning.patch
Patch2008: amp-8-aarch64-fix-xgene1-addrcost-table.patch
Patch2009: amp-9-aarch64-adapt-tuning-struct.patch
Patch2010: amp-10-add-pipelining-loads.patch
Patch2011: amp-11-uncse.patch
Patch2012: amp-12-loop-prefetcher-adapt-defaults.patch
Patch2013: amp-13-ssa-dont-crash-on-ssa-names.patch
Patch2014: amp-14-flag-funroll-more.patch
Patch2015: amp-15-bypass-alias.patch
Patch2016: amp-16-forced-creation-of-loop-preheader.patch
Patch2017: amp-17-noloopalias.patch

# Patches from gcc9 to support -flive-patching
Patch3001: gcc9-fipa-reference-addressable.patch
Patch3002: gcc9-fipa-stack-alignment.patch
Patch3003: gcc9-add-fomit-frame-pointer-to-test.patch
Patch3004: gcc9-extend-live-patching-option-handling.patch
Patch3005: gcc9-ipa-stack-alignment-386-test.patch

# Patches from gcc9 to support -fopt-info-inline
Patch3006: gcc9-opt-info-inline.patch

# Patches from gcc9 to fix PR47618
Patch3010: gcc9-pr47618.patch

# Patches from gcc9 to fix PR86057
Patch3011: gcc9-pr86057.patch

# Patches from gcc9 to fix PR85759 
Patch3012: gcc9-pr85759.patch 

# Patches from gcc10 to fix PR91971
Patch3013: gcc10-pr91971.patch

Patch3014: gcc-bpf.patch

%if 0%{?fedora} >= 29 || 0%{?rhel} > 7
BuildRequires: binutils >= 2.30-17
%else
BuildRequires: binutils >= 2.24
%endif
BuildRequires: zlib-devel, gettext, dejagnu, bison, flex, texinfo, sharutils
BuildRequires: binutils-bpf-unknown-none >= %{bpf_binutils_version}

# right now, we only have this for arm
#BuildRequires: glibc-arm-linux-gnu

# Make sure pthread.h doesn't contain __thread tokens
# Make sure glibc supports stack protector
# Make sure glibc supports DT_GNU_HASH
BuildRequires: glibc-devel >= 2.4.90-13
BuildRequires: elfutils-devel >= 0.147
BuildRequires: elfutils-libelf-devel >= 0.147
BuildRequires: gmp-devel >= 4.1.2-8, mpfr-devel >= 2.2.1, libmpc-devel >= 0.8.1
Provides: bundled(libiberty)

%description
BPF cross GNU C compiler collection.

###############################################################################
#
# Packages definition
#
###############################################################################

%package -n gcc-bpf-unknown-none
Summary: Cross GCC for bpf-unknown-none
BuildRequires: binutils-bpf-unknown-none >= %{bpf_binutils_version}
Requires: binutils-bpf-unknown-none >= %{bpf_binutils_version}
#%if 0%{?__isa_bits} == 64
#Requires: libisl.so.%{isl_libmajor}()(64bit)
#%else
#Requires: libisl.so.%{isl_libmajor}
#%endif
%description -n gcc-bpf-unknown-none
Cross-build BPF GNU C compiler.

###############################################################################
#
# Preparation
#
###############################################################################
%prep

%setup -q -n %{srcdir} -c
cd %{srcdir}

%patch0 -p0 -b .hack~
%patch2 -p0 -b .i386-libgomp~
%patch3 -p0 -b .sparc-config-detection~
%patch4 -p0 -b .libgomp-omp_h-multilib~
%patch5 -p0 -b .libtool-no-rpath~
%patch8 -p0 -b .no-add-needed~
%patch9 -p0 -b .foffload-default~
%patch10 -p0 -b .Wno-format-security~
%patch11 -p0 -b .rh1512529-aarch64~
%if 0%{?fedora} == 28
%patch12 -p0 -b .mcet~
%endif
%if 0%{?fedora} >= 29 || 0%{?rhel} > 7
%patch13 -p0 -b .rh1574936~
%patch14 -p0 -b .libgcc-hardened~
%endif
%patch15 -p0 -b .rh1670535~
%patch16 -p0 -b .pr85400~
%patch17 -p0 -b .libgomp-20190503~
%patch18 -p0 -b .pr86747~
%patch19 -p0 -b .libgomp-testsuite~
%patch20 -p0 -b .rh1711346~
%patch21 -p0 -b .rh1730380~
%patch22 -p0 -b .pr86098~
%patch23 -p0 -b .pr90139~
%patch24 -p0 -b .pr90756~

%patch30 -p0 -b .rh1668903-1~
%patch31 -p0 -b .rh1668903-2~
%patch32 -p0 -b .rh1668903-3~

%patch2001 -p1 -b .aarch64-retpoline~
%patch2002 -p1 -b .revert-richard~
%patch2003 -p1 -b .correct-shift~
%patch2004 -p1 -b .cpu-support~
%patch2005 -p1 -b .emag-xgene-cost~
%patch2006 -p1 -b .cost-table~
%patch2007 -p1 -b .prefetch-tuning~
%patch2008 -p1 -b .addrcost-table~
%patch2009 -p1 -b .tuning-struct~
%patch2010 -p1 -b .pipelining-loads~
%patch2011 -p1 -b .uncse~
%patch2012 -p1 -b .loop-prefetcher~
%patch2013 -p1 -b .ssa-dont-crash~
%patch2014 -p1 -b .flag-funroll-more~
%patch2015 -p1 -b .tree-vect-slp~
%patch2016 -p1 -b .forced-creation~
%patch2017 -p1 -b .noloopalias~

%patch3001 -p1 -b  .fipa-reference-addressable~
%patch3002 -p1 -b  .fipa-stack-alignment~
%patch3003 -p1 -b  .add-fomit-frame-pointer-to-test~
%patch3004 -p1 -b  .extend-live-patching-option-handling~
%patch3005 -p1 -b  .ipa-stack-alignment-386-test~

%patch3006 -p1 -b  .opt-info-inline~

%patch3010 -p1 -b  .pr47618~

%patch3011 -p1 -b  .pr86057~
%patch3012 -p1 -b  .pr85759~
%patch3013 -p1 -b  .pr91971~
%patch3014 -p2 -b  .bpf

echo 'BPF GCC cross %{version}-%{gcc_release}' > gcc/DEV-PHASE

echo 'TM_H += $(srcdir)/config/rs6000/rs6000-modes.h' >> gcc/config/rs6000/t-rs6000

./contrib/gcc_update --touch

LC_ALL=C sed -i -e 's/\xa0/ /' gcc/doc/options.texi

sed -i -e 's/Common Driver Var(flag_report_bug)/& Init(1)/' gcc/common.opt

# This test causes fork failures, because it spawns way too many threads
rm -f gcc/testsuite/go.test/test/chan/goroutines.go

cd ..

###############################################################################
#
# Build
#
###############################################################################
%build

%global builddir %{_builddir}/%{srcdir}

# Undo the broken autoconf change in recent Fedora versions
export CONFIG_SITE=NONE

CC=gcc
CXX=g++
OPT_FLAGS=`echo %{optflags}|sed -e 's/\(-Wp,\)\?-D_FORTIFY_SOURCE=[12]//g'`
OPT_FLAGS=`echo $OPT_FLAGS|sed -e 's/-m64//g;s/-m32//g;s/-m31//g'`
OPT_FLAGS=`echo $OPT_FLAGS|sed -e 's/-mfpmath=sse/-mfpmath=sse -msse2/g'`
OPT_FLAGS=`echo $OPT_FLAGS|sed -e 's/ -pipe / /g'`
%ifarch sparc
OPT_FLAGS=`echo $OPT_FLAGS|sed -e 's/-mcpu=ultrasparc/-mtune=ultrasparc/g;s/-mcpu=v[78]//g'`
%endif
%ifarch %{ix86}
OPT_FLAGS=`echo $OPT_FLAGS|sed -e 's/-march=i.86//g'`
%endif
OPT_FLAGS=`echo $OPT_FLAGS|sed -e 's/-Werror=format-security/-Wformat-security/g'`
OPT_FLAGS=`echo "$OPT_FLAGS" | sed -e 's/[[:blank:]]\+/ /g'`
case "$OPT_FLAGS" in
  *-fasynchronous-unwind-tables*)
    sed -i -e 's/-fno-exceptions /-fno-exceptions -fno-asynchronous-unwind-tables /' \
      %{srcdir}/libgcc/Makefile.in
    ;;
esac

#
# Configure the compiler
#
cd %{builddir}

echo "=== CONFIGURING bpf-unknown-none"

arch=bpf-unknown-none
prefix=$arch-
build_dir=bpf-unknown-none
target=bpf-unknown-none

echo $arch: target is $target

CONFIG_FLAGS="--disable-shared"

mkdir $build_dir
cd $build_dir

# We could optimize the cross builds size by --enable-shared but the produced
# binaries may be less convenient in the embedded environment.
CC="$CC" \
CXX="$CXX" \
CFLAGS="$OPT_FLAGS" \
CXXFLAGS="`echo " $OPT_FLAGS " | sed 's/ -Wall / /g;s/ -fexceptions / /g' \
	  | sed 's/ -Wformat-security / -Wformat -Wformat-security /'`" \
CFLAGS_FOR_TARGET="-g -O2 -Wall -fexceptions" \
AR_FOR_TARGET=%{_bindir}/$arch-ar \
AS_FOR_TARGET=%{_bindir}/$arch-as \
DLLTOOL_FOR_TARGET=%{_bindir}/$arch-dlltool \
LD_FOR_TARGET=%{_bindir}/$arch-ld \
NM_FOR_TARGET=%{_bindir}/$arch-nm \
OBJDUMP_FOR_TARGET=%{_bindir}/$arch-objdump \
RANLIB_FOR_TARGET=%{_bindir}/$arch-ranlib \
READELF_FOR_TARGET=%{_bindir}/$arch-readelf \
STRIP_FOR_TARGET=%{_bindir}/$arch-strip \
WINDRES_FOR_TARGET=%{_bindir}/$arch-windres \
WINDMC_FOR_TARGET=%{_bindir}/$arch-windmc \
LDFLAGS='-Wl,-z,relro ' \
../%{srcdir}/configure \
	--bindir=%{_bindir} \
	--build=%{_target_platform} \
	--datadir=%{_datadir} \
	--disable-decimal-float \
	--disable-dependency-tracking \
	--disable-gold \
	--disable-libgcj \
	--disable-libgomp \
	--disable-libmpx \
	--disable-libquadmath \
	--disable-libssp \
	--disable-libunwind-exceptions \
	--disable-silent-rules \
	--disable-sjlj-exceptions \
	--disable-threads \
	--with-ld=/usr/bin/$arch-ld \
	--enable-__cxa_atexit \
	--enable-checking=release \
	--enable-gnu-unique-object \
	--enable-initfini-array \
	--enable-languages=c \
	--enable-linker-build-id \
	--enable-lto \
	--enable-nls \
	--enable-obsolete \
	--enable-plugin \
	--enable-targets=all \
	--exec-prefix=%{_exec_prefix} \
	--host=%{_target_platform} \
	--includedir=%{_includedir} \
	--infodir=%{_infodir} \
	--libexecdir=%{_libexecdir} \
	--localstatedir=%{_localstatedir} \
	--mandir=%{_mandir} \
	--prefix=%{_prefix} \
	--program-prefix=$prefix \
	--sbindir=%{_sbindir} \
	--sharedstatedir=%{_sharedstatedir} \
	--sysconfdir=%{_sysconfdir} \
	--target=$target \
	--with-bugurl=http://bugzilla.redhat.com/bugzilla/ \
	--with-gcc-major-version-only \
	--without-isl \
	--with-newlib \
	--with-plugin-ld=%{_bindir}/$arch-ld \
	--with-sysroot=%{_prefix}/$arch/sys-root \
	--with-system-libunwind \
	--with-system-zlib \
	--without-headers \
	$CONFIG_FLAGS
cd ..

#
# Build the compiler
#

echo "=== BUILDING bpf-unknown-none"

arch=bpf-unknown-none
build_dir=bpf-unknown-none

BUILD_FLAGS=

AR_FOR_TARGET=%{_bindir}/$arch-ar \
AS_FOR_TARGET=%{_bindir}/$arch-as \
DLLTOOL_FOR_TARGET=%{_bindir}/$arch-dlltool \
LD_FOR_TARGET=%{_bindir}/$arch-ld \
NM_FOR_TARGET=%{_bindir}/$arch-nm \
OBJDUMP_FOR_TARGET=%{_bindir}/$arch-objdump \
RANLIB_FOR_TARGET=%{_bindir}/$arch-ranlib \
READELF_FOR_TARGET=%{_bindir}/$arch-readelf \
STRIP_FOR_TARGET=%{_bindir}/$arch-strip \
WINDRES_FOR_TARGET=%{_bindir}/$arch-windres \
WINDMC_FOR_TARGET=%{_bindir}/$arch-windmc \
make -C $build_dir %{_smp_mflags} tooldir=%{_prefix} $BUILD_FLAGS all-gcc

echo "=== BUILDING LIBGCC bpf-unknown-none"
make -C $build_dir tooldir=%{_prefix} all-target-libgcc

#
# Installation
#
%install
rm -rf %{buildroot}

mkdir -p %{buildroot}%{_prefix}/bpf-unknown-none/sys-root
echo "=== INSTALLING bpf-unknown-none"

arch=bpf-unknown-none
cpu=${arch%%%%-*}

make -C $arch DESTDIR=%{buildroot} install-gcc install-target-libgcc

# Drop the documentation.
rm -rf %{buildroot}%{_infodir}

# Remove binaries we will not be including, so that they don't end up in
# bpf-gcc-debuginfo
rm -f %{buildroot}%{_libdir}/{libffi*,libiberty.a}
rm -f %{buildroot}%{_libexecdir}/gcc/*/%{gcc_major}/install-tools/{mkheaders,fixincl}
rm -f %{buildroot}%{_prefix}/bin/*-gcc-%{gcc_major} || :
rmdir  %{buildroot}%{_includedir}

find %{buildroot}%{_datadir} -name gcc.mo |
while read x
do
    y=`dirname $x`
    mv $x $y/bpf-gcc.mo
done

%find_lang bpf-gcc

gzip %{buildroot}%{_mandir}/man1/*.1
rm %{buildroot}%{_mandir}/man7/*.7
rmdir %{buildroot}%{_mandir}/man7

# # Add manpages the additional symlink-only targets
# %if %{build_powerpc}%{build_powerpc64}%{build_powerpc64le}
# for i in powerpc*
# do
#     ln -s $i ppc${i#powerpc}
# done
# %endif

#cd -

#
# Build the file lists
#

arch=bpf-unknown-none
cpu=${arch%%%%-*}

(
  echo '%{_bindir}/'$arch'*-cpp'
  echo '%{_bindir}/'$arch'*-gcc'
  echo '%{_bindir}/'$arch'*-gcc-ar'
  echo '%{_bindir}/'$arch'*-gcc-nm'
  echo '%{_bindir}/'$arch'*-gcc-ranlib'
  echo '%{_bindir}/'$arch'*-gcov*'
  echo '%{_mandir}/man1/'$arch'*-cpp*'
  echo '%{_mandir}/man1/'$arch'*-gcc*'
  echo '%{_mandir}/man1/'$arch'*-gcov*'
  echo '/usr/lib/gcc/'$cpu'-*/'
  echo '%{_libexecdir}/gcc/'$cpu'*/*/cc1'
  echo '%{_libexecdir}/gcc/'$cpu'*/*/collect2'
  echo '%{_libexecdir}/gcc/'$cpu'*/*/[abd-z]*'
  echo %{_prefix}/$arch/sys-root
) >files.$arch

%global __ar_no_strip $RPM_BUILD_DIR/%{srcdir}/ar-no-strip
cat >%{__ar_no_strip} <<EOF
#!/bin/bash
f=\$2
if [ \${f##*/} = libgcc.a -o \${f##*/} = libgcov.a ]
then
	:
else
	%{__strip} \$*
fi
EOF
chmod +x %{__ar_no_strip}
%undefine __strip
%global __strip %{__ar_no_strip}

###############################################################################
#
# Filesets
#
###############################################################################

%files -n gcc-bpf-unknown-none -f files.bpf-unknown-none -f bpf-gcc.lang
%doc %{srcdir}/COPYING*
%doc %{srcdir}/README

%changelog
* Fri May 01 2020 Jose E. Marchesi <jose.marchesi@oracle.com> - 8.3.1-3
- Avoid generation of funcalls for one_cmpl{si,di}2 insns in BPF.
- Reviewed-by: Elena Zannoni <elena.zannoni@oracle.com>

* Tue Apr 21 2020 Jose E. Marchesi <jose.marchesi@oracle.com> - 8.3.1-2
- Support for the -mxbpf command line option.
- Do not save/restore callee-saved registers in function
  prologues/epilogues unless generating xBPF.
- Reviewed-by: Elena Zannoni <elena.zannoni@oracle.com>

* Wed Mar 11 2020 Jose E. Marchesi <jose.marchesi@oracle.com> - 8.3.1-1
- New package.
